(* -*- coding: utf-8 -*- *)
(*
  Copyright © 2017 Thibaut Girka / Mitsubishi Electric R&D Centre Europe / Institut de Recherche en Informatique Fondamentale
  Copyright © 2017 Yann Régis-Gianas

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. Neither the name of the University nor the names of its contributors
     may be used to endorse or promote products derived from this software
     without specific prior written permission.

  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  SUCH DAMAGE.
 *)
(** General definitions about reduction traces relations.

    Relations between traces are represented by binary predicates. 

*)

Require Import Common.
Require Import ProgrammingLanguage.
Require Import ReductionTrace.

(** A relation between traces is represented by a binary predicate. *)
Definition trace_relation (L1 L2 : language) :=
  coplist (configuration L1) -> coplist (configuration L2) -> Prop.

Definition compose_trace_relations_externally
  (L1 L2 L3 : language)
  (Δ12 : trace_relation L1 L2) (Δ23 : trace_relation L2 L3)
: trace_relation L1 L3
:=
  λ tr1 tr3, ∃ tr2, Δ12 tr1 tr2 /\ Δ23 tr2 tr3.

CoInductive gamma_correlated {L₁ L₂ : language}
            (γ : configuration L₁ → configuration L₂ → Prop)
  : coplist (configuration L₁) → coplist (configuration L₂) → Prop :=
| TerminatedCorrelated:
    ∀ c₁ c₂,
      γ c₁ c₂ →
      gamma_correlated γ (CSingle _ c₁) (CSingle _ c₂)
| StepCorrelated:
    ∀ tr₁ tr₂ n₁ n₂ tr₁' tr₂',
      γ (first tr₁) (first tr₂) →
      skipn n₁ tr₁ = Some tr₁' →
      skipn n₂ tr₂ = Some tr₂' →
      n₁ + n₂ > 0 →
      gamma_correlated γ tr₁' tr₂' →
      gamma_correlated γ tr₁ tr₂.